// Model.cpp: implementation of the CModel class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Model.h"

#include <stdio.h>

#include "Name.h"
#include "State.h"

#define kMaxCarID 200
#define kMaxTracks 2
#define kMaxLanes 4

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CModel::CModel() {
  mName = new CName();
  mState = new CState();

  mCurrentState = mState;

  mLogicalClock = 0;

  mAutosaveFilename = 0;
  mAutosaveFilename = new char[1024];
  time_t theTime = time(0);
  sprintf(mAutosaveFilename, "autosave_%d.txt", theTime);
}

CModel::~CModel() {
  delete mName;
  delete mState;

  if (mAutosaveFilename) {
    delete mAutosaveFilename;
  }
}

//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////

int CModel::getMaxCarID() const {
  return kMaxCarID;
}

void CModel::autosave() {
  CState* theState = mCurrentState->clone();
  mCurrentState = theState;

  // implement autosave
  if (mAutosaveFilename) {
    FILE* theFile;
    theFile = fopen(mAutosaveFilename, "w+t");
    save(theFile);
    fclose(theFile);
  }
}

bool CModel::trackSelect(int pTrack, int pLane, int pCarID) {
  if (!checkTrack(pTrack)) {
    return false;
  }
  if (!checkLane(pLane)) {
    return false;
  }
  
  if (!isIDRaceable(pCarID)) {
    return false;
  }

  // id is not raceable anymore
  clearIDRaceable(pCarID);

  // put car id into selection
  mCurrentState->mSelection[pTrack][pLane] = pCarID;
  
  return false;
}

int CModel::trackGetSelected(int pTrack, int pLane) {
  if (!checkTrack(pTrack)) {
    return -1;
  }
  if (!checkLane(pLane)) {
    return -1;
  }

  // get id on the track
  return mCurrentState->mSelection[pTrack][pLane];
}

char* CModel::trackGetSelectedStr(int pTrack, int pLane) {
  return toString(trackGetSelected(pTrack, pLane));
}

int CModel::trackGetNumSelected(int pTrack) {
  if (!checkTrack(pTrack)) {
    return -1;
  }

  int theCounter = 0;

  for (int i = 0; i < kMaxLanes; i++) {
    if (mCurrentState->mSelection[pTrack][i] > -1) {
      theCounter++;
    }
  }

  return theCounter;
}

int CModel::trackUnselect(int pTrack, int pLane) {
  // get id on the track
  int theID = trackGetSelected(pTrack, pLane);

  // unset id on track
  mCurrentState->mSelection[pTrack][pLane] = -1;

  // set raceable flag for theID
  setIDRaceable(theID);

  return theID;
}

void CModel::trackRandomSelect(int pTrack, int pHoldLane0, int pHoldLane1, int pHoldLane2, int pHoldLane3) {
  if (!checkTrack(pTrack)) {
    return;
  }

  int theLanesLeft = 0;

  // unselect lanes
  if (pHoldLane0 == 0) {
    trackUnselect(pTrack, 0);
    theLanesLeft++;
  }

  if (pHoldLane1 == 0) {
    trackUnselect(pTrack, 1);
    theLanesLeft++;
  }

  if (pHoldLane2 == 0) {
    trackUnselect(pTrack, 2);
    theLanesLeft++;
  }

  if (pHoldLane3 == 0) {
    trackUnselect(pTrack, 3);
    theLanesLeft++;
  }

  // randomly select up to kMaxLanes different items from the pool
  // - rely on probabilistic positive selections as not all random numbers
  //   correspond to car IDs
  int theCounter = 0;
  int theLanesSelected = 0;

  while (theCounter < kMaxLanes) {
    // don't select for some lanes
    if (((theCounter == 0) && (pHoldLane0 != 0)) ||
        ((theCounter == 1) && (pHoldLane1 != 0)) ||
        ((theCounter == 2) && (pHoldLane2 != 0)) ||
        ((theCounter == 3) && (pHoldLane3 != 0))) {
      theCounter++;
      continue;
    }

    // select id and place in lane
    int theCarID = rand() % kMaxCarID;

    if (isIDRaceable(theCarID)) {
      trackSelect(pTrack, theCounter, theCarID);
      theCounter++;
      theLanesSelected++;
      theLanesLeft--;
    } // if (isIDRaceable(theCarID))

    // don't select anymore if we have either 0 or 2 left
    // - 0 because there won't be anything left after this
    // - 2 because we don't want to race only one car against itself.
    //   - but also make sure that we're not running only one car right now
    // - this isn't a perfect solution, but there really isn't a perfect solution
    //   - there are, however, better solutions, I think
    int theRaceableInPool = getRaceableInPool();
    if ((theRaceableInPool == 0) ||
        ((theRaceableInPool <= 2) && (theLanesSelected > 1) && (theLanesLeft < theRaceableInPool))) {
      for (int i = theCounter; i < kMaxLanes; i++) {
        trackSelect(pTrack, i, -1);
      }

      // don't continue anymore because we can't select anymore
      break;
    } // if (theRaceableInPool == 0)
  }
}

bool CModel::raceSetNumber(int pTrack, int pLane, int pCarID) {
  if (!checkTrack(pTrack)) {
    return false;
  }
  if (!checkLane(pLane)) {
    return false;
  }

  if (!isIDRaceable(pCarID)) {
    return false;
  }

  // id is not raceable anymore
  clearIDRaceable(pCarID);
  
  // assign numbers to current race
  mCurrentState->mTrackScore[pTrack][pLane].mID = pCarID;

  return true;
}

int CModel::raceGetNumber(int pTrack, int pLane) {
  if (!checkTrack(pTrack)) {
    return -1;
  }
  if (!checkLane(pLane)) {
    return -1;
  }

  // get car id
  return mCurrentState->mTrackScore[pTrack][pLane].mID;
}

char* CModel::raceGetNumberStr(int pTrack, int pLane) {
  return toString(raceGetNumber(pTrack, pLane));
}

int CModel::raceGetNumSelected(int pTrack) {
  if (!checkTrack(pTrack)) {
    return -1;
  }

  int theCounter = 0;

  for (int i = 0; i < kMaxLanes; i++) {
    if (mCurrentState->mTrackScore[pTrack][i].mID > -1) {
      theCounter++;
    }
  }

  return theCounter;
}

int CModel::raceUnsetNumber(int pTrack, int pLane) {
  // get id on race
  int theID = raceGetNumber(pTrack, pLane);

  // unset id on race
  mCurrentState->mTrackScore[pTrack][pLane] = CState::sIDScore();

  // set raceable flag for theID
  setIDRaceable(theID);

  return theID;
}

void CModel::transferSelectionToRace(int pTrack) {
  if (!checkTrack(pTrack)) {
    return;
  }

  // transfer selection list to race and reset race results
  for (int i = 0; i < kMaxLanes; i++) {
    // unset previous id from race
    raceUnsetNumber(pTrack, i);

    // get id from selection and unselect
    int theID = trackUnselect(pTrack, i);

    // set id into race
    raceSetNumber(pTrack, i, theID);
  }
}

bool CModel::raceSetResult(int pTrack, int pLane, int pPlace) {
  if (!checkTrack(pTrack)) {
    return false;
  }
  if (!checkLane(pLane)) {
    return false;
  }
  
  // set race results
  mCurrentState->mTrackScore[pTrack][pLane].mScore = pPlace;

  return true;
}

bool CModel::raceSetResult(int pTrack, int pLane) {
  if (!checkTrack(pTrack)) {
    return false;
  }
  if (!checkLane(pLane)) {
    return false;
  }

  // get the greatest place already set
  int theMaxPlace = -1;
  for (int i = 0; i < kMaxLanes; i++) {
    int theTrackScore = mCurrentState->mTrackScore[pTrack][i].mScore;
    if (theTrackScore > theMaxPlace) {
      theMaxPlace = theTrackScore;
    }
  }

  // advance to the next place
  int theNextPlace;
  if (theMaxPlace == -1) {
    theNextPlace = 1;
  } else if (theMaxPlace < kMaxLanes) {
    // if the track/lane combination is already set, then don't set it again
    if (mCurrentState->mTrackScore[pTrack][pLane].mScore > -1) {
      return false;
    }

    theNextPlace = theMaxPlace + 1;
  } else {
    // reset all places to -1 (unset)
    for (int i = 0; i < kMaxLanes; i++) {
      mCurrentState->mTrackScore[pTrack][i].mScore = -1;
    }

    // we're done...
    return true;
  }

  // set race results
  mCurrentState->mTrackScore[pTrack][pLane].mScore = theNextPlace;

  return true;
}

int CModel::raceGetResult(int pTrack, int pLane) {
  if (!checkTrack(pTrack)) {
    return -1;
  }
  if (!checkLane(pLane)) {
    return -1;
  }
  
  // set race results
  if (mCurrentState->mTrackScore[pTrack][pLane].mID > -1) {
    return mCurrentState->mTrackScore[pTrack][pLane].mScore;
  } else {
    return -1;
  }
}

char* CModel::raceGetResultStr(int pTrack, int pLane) {
  return toString2(raceGetResult(pTrack, pLane));
}

void CModel::raceCompactResults(int pTrack) {
  if (!checkTrack(pTrack)) {
    return;
  }

  for (int i = 0; i < kMaxLanes; i++) {
    int thePlace = raceGetResult(pTrack, i);

    // handle each lane separately
    if (thePlace > 1) {
      // find if immediate precedent exists
      bool thePrecedentExists = false;

      for (int j = 0; j < kMaxLanes; j++) {
        if (i == j) {
          continue;
        }

        if (raceGetResult(pTrack, j) == (thePlace - 1)) {
          thePrecedentExists = true;
          break;
        }
      } // for (j < kMaxLanes)

      // if precedent doesn't exist, decrease myself
      if (!thePrecedentExists) {
        raceSetResult(pTrack, i, thePlace - 1);

        // also start the i search over again
        // - -1 to account for the i++ iterative
        i = -1;
      }
    } // if (thePlace > 1)
  } // for (i < kMaxLanes)
}

char* CModel::getNameForID(int pCarID) {
  strcpy(mStrBuffer, mName->get(pCarID).c_str());
  return mStrBuffer;
}

void CModel::setNameForID(int pCarID, const char* pName) {
  mName->change(pCarID, pName);
}

void CModel::raceClearResults(int pTrack) {
  if (!checkTrack(pTrack)) {
    return;
  }

  // clear current race results
  autosave();
  
  for (int i = 0; i < kMaxLanes; i++) {
    mCurrentState->mTrackScore[pTrack][i].mScore = 0;
  }
}

void CModel::commitResults(int pTrack) {
  if (!checkTrack(pTrack)) {
    return;
  }

  int theLogicalClock = getLogicalClock();

  // move mTrackScore to mScore
  autosave();

  for (CState::tIDScoreMAP::iterator theIter = mCurrentState->mTrackScore[pTrack].begin(); theIter != mCurrentState->mTrackScore[pTrack].end(); theIter++) {
    int theID = theIter->second.mID;
    int theScore = theIter->second.mScore;

    if (theID < 0) {
      continue;
    }

    // record score
    mCurrentState->mScore[theID].mOrder = theLogicalClock - theScore;
    mCurrentState->mScore[theID].mList.push_back(theScore);

    // clear score in track results
    theIter->second = CState::sIDScore();
  }
}

bool CModel::isIDRaceable(int pCarID) {
  if (mCurrentState->mPool.find(pCarID) == mCurrentState->mPool.end()) {
    return false;
  }

  return (mCurrentState->mPool[pCarID] == 0);
}

void CModel::setIDRaceable(int pCarID) {
  if (pCarID > -1) {
    mCurrentState->mPool[pCarID] = 0;
  }
}

void CModel::clearIDRaceable(int pCarID) {
  if (mCurrentState->mPool.find(pCarID) == mCurrentState->mPool.end()) {
    return;
  }

  mCurrentState->mPool[pCarID] = 1;
}

int CModel::getRaceableInPool() {
  int theCounter = 0;

  for (CState::tIntIntMAP::iterator theIter = mCurrentState->mPool.begin(); theIter != mCurrentState->mPool.end(); theIter++) {
    if (isIDRaceable(theIter->first)) {
      theCounter++;
    }
  }

  return theCounter;
}

void CModel::setAllInPoolRaceable() {
  autosave();

  for (int i = 0; i < kMaxTracks; i++) {
    for (int j = 0; j < kMaxLanes; j++) {
      mCurrentState->mSelection[i][j] = -1;
      mCurrentState->mTrackScore[i][j] = CState::sIDScore(-1, 0);
    }
  }
  
  for (CState::tIntIntMAP::iterator theIter = mCurrentState->mPool.begin(); theIter != mCurrentState->mPool.end(); theIter++) {
    theIter->second = 0;
  }
}

CModel::tIntLIST CModel::getIDsInPool() {
  tIntLIST theRet;

  for (CState::tIntIntMAP::const_iterator theIter = mCurrentState->mPool.begin(); theIter != mCurrentState->mPool.end(); theIter++) {
    theRet.push_back(theIter->first);
  }

  return theRet;
}

bool CModel::inPool(int pID) const {
  return (mCurrentState->mPool.find(pID) != mCurrentState->mPool.end());
}

void CModel::addToPool(int pID) {
  if (mCurrentState->mPool.find(pID) == mCurrentState->mPool.end()) {
    mCurrentState->mPool[pID] = 0;
  }
}

void CModel::removeFromPool(int pID) {
  if (mCurrentState->mPool.find(pID) != mCurrentState->mPool.end()) {
    mCurrentState->mPool.erase(pID);
  }
}

CModel::tIntLIST CModel::getIDsWithScores() {
  tIntIntMAP theTemp;
  tIntLIST theRet;

  { // re-sort list according to the map
    for (CState::tIntIntScoreListMAP::iterator theIter = mCurrentState->mScore.begin(); theIter != mCurrentState->mScore.end(); theIter++) {
      const CState::tIntLIST *theList = &(theIter->second.mList);
      if (theList->size() > 0) {
        // make sure we don't have a duplicate order
        int theOrder = theIter->second.mOrder;
        while (theTemp.find(theOrder) != theTemp.end()) {
          theOrder -= 1;
        }
        theTemp[theOrder] = theIter->first;

        // write-back makes things more efficient, and order isn't critical anyway
        theIter->second.mOrder = theOrder;      
      }
    }
  }

  { // populate return list
    for (tIntIntMAP::const_iterator theIter = theTemp.begin(); theIter != theTemp.end(); theIter++) {
      theRet.push_front(theIter->second);
    }
  }

  return theRet;
}

CModel::tIntLIST CModel::getScoresForID(int pID) {
  tIntLIST theRet;

  if (mCurrentState->mScore.find(pID) != mCurrentState->mScore.end()) {
    theRet = mCurrentState->mScore[pID].mList;
  }

  return theRet;
}

void CModel::eliminate(int pQuantity, int pScore) {
  autosave();

  // clear selection and track scores
  for (int i = 0; i < kMaxTracks; i++) {
    for (int j = 0; j < kMaxLanes; j++) {
      mCurrentState->mSelection[i][j] = -1;
      mCurrentState->mTrackScore[i][j] = CState::sIDScore(-1, 0);
    }
  }
  
  // eliminate
  tIntLIST theKeep;
  tIntLIST theIDsWithScores = getIDsWithScores();
  {
    for (tIntLIST::const_iterator theIter = theIDsWithScores.begin(); theIter != theIDsWithScores.end(); theIter++) {
      int theID = *theIter;
      tIntLIST theScores = getScoresForID(theID);

      int theCounter = 0;
      for (tIntLIST::const_iterator theIter2 = theScores.begin(); theIter2 != theScores.end(); theIter2++) {
        int theScore = *theIter2;
        if ((theScore >= 1) && (theScore <= pScore)) {
          theCounter++;
        }
      }

      if (theCounter >= pQuantity) {
        // made the cut
        theKeep.push_back(theID);
      }
    }
  }

  // clear scores
  mCurrentState->mScore.clear();

  { // repopulate pool
    mCurrentState->mPool.clear();
    for (tIntLIST::const_iterator theIter = theKeep.begin(); theIter != theKeep.end(); theIter++) {
      int theID = *theIter;
      mCurrentState->mPool[theID] = 0;
    }
  }

}

const CState* CModel::getCurrentState() const {
  return mCurrentState;
}

bool CModel::canUndo() {
  return mCurrentState->canUndo();
}

bool CModel::canRedo() {
  return mCurrentState->canRedo();
}

void CModel::undo() {
  CState* theState = mCurrentState->undo();
  if (theState) {
    mCurrentState = theState;
  }
}

void CModel::redo() {
  CState* theState = mCurrentState->redo();
  if (theState) {
    mCurrentState = theState;
  }
}

void CModel::load(FILE* pFile) {
  mName->load(pFile);
  mState->load(pFile);

  // default initialize pool if no state
  if (mState->isClear()) {
    mState->initPoolWithCName(*mName);
  }

  // update mCurrentState
  mCurrentState = mState->lastState();
}

void CModel::save(FILE* pFile) {
  mName->save(pFile);
  mCurrentState->save(pFile);
}

//////////////////////////////////////////////////////////////////////
// protected
//////////////////////////////////////////////////////////////////////

bool CModel::checkTrack(int pTrack) {
  return ((pTrack >= 0) && (pTrack < kMaxTracks));
}

bool CModel::checkLane(int pLane) {
  return ((pLane >= 0) && (pLane < kMaxLanes));
}

char* CModel::toString(int pInt) {
  if (pInt > -1) {
    sprintf(mStrBuffer, "%02d", pInt);
    return mStrBuffer;
  } else {
    return "--";
  }
}

char* CModel::toString2(int pInt) {
  if (pInt > -1) {
    sprintf(mStrBuffer, "%d", pInt);
    return mStrBuffer;
  } else {
    return "--";
  }
}

int CModel::getLogicalClock() {
  mLogicalClock++;
  return mLogicalClock * 10;
}

//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////

void CModel::initWithTestData() {
  mName->add(0,  "zero");
  mName->add(1,  "alpha");
  mName->add(2,  "bravo");
  mName->add(3,  "charlie");
  mName->add(4,  "delta");
  mName->add(5,  "echo");
  mName->add(6,  "foxtrot");
  mName->add(7,  "gulf");
  mName->add(8,  "hotel");
  mName->add(9,  "india");
  mName->add(10, "juliet");
  mName->add(11, "kilo");
  mName->add(12, "lima");
  mName->add(13, "mike");
  mName->add(14, "november");
  mName->add(15, "oscar");
  mName->add(16, "peru");
  mName->add(17, "quebec");
  mName->add(18, "romeo");
  mName->add(19, "sierra");
  mName->add(20, "tango");
  mName->add(21, "uniform");
  mName->add(22, "victor");
  mName->add(23, "whiskey");
  mName->add(24, "x-ray");
  mName->add(25, "yankee");
  mName->add(26, "zulu");

  mName->add(101, "genesis");
  mName->add(102, "exodus");
  mName->add(103, "leviticus");
  mName->add(104, "numbers");
  mName->add(105, "deuteronomy");
  mName->add(106, "joshua");
  mName->add(107, "judges");
  mName->add(108, "ruth");
  mName->add(109, "1 samuel");
  mName->add(110, "2 samuel");
  mName->add(111, "1 kings");
  mName->add(112, "2 kings");
  mName->add(113, "1 chronicles");
  mName->add(114, "2 chronicles");
  mName->add(115, "ezra");
  mName->add(116, "nehemiah");
  mName->add(117, "esther");
  mName->add(118, "job");
  mName->add(119, "psalms");
  mName->add(120, "proverbs");
  mName->add(121, "ecclesiastes");
  mName->add(122, "song of songs");
  mName->add(123, "isaiah");
  mName->add(124, "jeremiah");
  mName->add(125, "lamentations");
  mName->add(126, "ezekiel");
  mName->add(127, "daniel");
  mName->add(128, "hosea");
  mName->add(129, "joel");
  mName->add(130, "amos");
  mName->add(131, "obadiah");
  mName->add(132, "jonah");
  mName->add(133, "micah");
  mName->add(134, "nahum");
  mName->add(135, "habakkuk");
  mName->add(136, "zephaniah");
  mName->add(137, "haggai");
  mName->add(138, "zechariah");
  mName->add(139, "malachi");
  mName->add(140, "matthew");
  mName->add(141, "mark");
  mName->add(142, "luke");
  mName->add(143, "john");
  mName->add(144, "acts");
  mName->add(145, "romans");
  mName->add(146, "1 corinthians");
  mName->add(147, "2 corinthians");
  mName->add(148, "galatians");
  mName->add(149, "ephesians");
  mName->add(150, "philippians");
  mName->add(151, "colossians");
  mName->add(152, "1 thessalonians");
  mName->add(153, "2 thessalonians");
  mName->add(154, "1 timothy");
  mName->add(155, "2 timothy");
  mName->add(156, "titus");
  mName->add(157, "philemon");
  mName->add(158, "hebrews");
  mName->add(159, "james");
  mName->add(160, "1 peter");
  mName->add(161, "2 peter");
  mName->add(162, "1 john");
  mName->add(163, "2 john");
  mName->add(164, "3 john");
  mName->add(165, "jude");
  mName->add(166, "revelation");

  mState->initPoolWithCName(*mName);
  mCurrentState = mState->lastState();

  return;

  trackRandomSelect(0);
  trackRandomSelect(1);

  transferSelectionToRace(0);
  transferSelectionToRace(1);

  trackRandomSelect(0);
  trackRandomSelect(1);

}
