// Model.h: interface for the CModel class.
// - interface to the gui
// - also controls underlaying logic
//
//////////////////////////////////////////////////////////////////////

#ifndef __model_h__
#define __model_h__

#ifdef _MSC_VER
  // microsoft vc++ specific
  #pragma warning(disable:4786)
#endif

#include <list>
#include <map>

using namespace std;

class CName;
class CState;

class CModel {
  public:
    typedef list<int> tIntLIST;
    typedef map<int, int> tIntIntMAP;

	  CModel();
	  virtual ~CModel();

    // max car id?
    int getMaxCarID() const;

    // autosave
    void autosave();

    // track select
    bool trackSelect(int pTrack, int pLane, int pCarID);

    // track get selected
    int trackGetSelected(int pTrack, int pLane);
    char* trackGetSelectedStr(int pTrack, int pLane);

    // track get number selected
    int trackGetNumSelected(int pTrack);

    // track unselect
    int trackUnselect(int pTrack, int pLane);

    // track random reselect
    // - if pHoldLaneX != 0 then don't use the lane
    void trackRandomSelect(int pTrack, int pHoldLane0 = 0, int pHoldLane1 = 0, int pHoldLane2 = 0, int pHoldLane3 = 0);

    // track set race numbers
    bool raceSetNumber(int pTrack, int pLane, int pCarID);

    // track get race numbers
    int raceGetNumber(int pTrack, int pLane);
    char* raceGetNumberStr(int pTrack, int pLane);
    
    // race get number selected
    int raceGetNumSelected(int pTrack);

    // track unset race numbers
    int raceUnsetNumber(int pTrack, int pLane);

    // transfer number to race
    void transferSelectionToRace(int pTrack);

    // track set race results
    // - first version explicit
    // - second version: implicit, adds new place to maximum place already set
    bool raceSetResult(int pTrack, int pLane, int pPlace);
    bool raceSetResult(int pTrack, int pLane);

    // track get race results
    int raceGetResult(int pTrack, int pLane);
    char* raceGetResultStr(int pTrack, int pLane);

    // track compact race results such that there is never a higher place missing
    void raceCompactResults(int pTrack);

    // get name for id
    char* getNameForID(int pCarID);

    // set name for id
    void setNameForID(int pCarID, const char* pName);

    // clear race results
    // - autosaves
    void raceClearResults(int pTrack);

    // track commit race results
    // - autosaves
    void commitResults(int pTrack);

    // is the id raceable in the pool?
    bool isIDRaceable(int pCarID);

    // set the raceable flag for id in pool
    void setIDRaceable(int pCarID);

    // clear raceable flag for id in pool
    void clearIDRaceable(int pCarID);

    // counts number of raceable in pool
    int getRaceableInPool();

    // set everything in the pool raceable
    // - also removes all entries from selection and race lists
    // - autosaves
    void setAllInPoolRaceable();


    // get a list of ids in the pool
    tIntLIST getIDsInPool();

    // is an id in the pool?
    bool inPool(int pID) const;

    // add id to pool
    void addToPool(int pID);

    // remove id from pool
    void removeFromPool(int pID);

    // get a list of ids with scores
    tIntLIST getIDsWithScores();

    // get a list of scores for each id
    tIntLIST getScoresForID(int pID);


    // eliminate
    // - eliminate all ids that do not satisfy at least pQuantity of at least pScore
    // - clears scores
    // - autosaves
    void eliminate(int pQuantity, int pScore);


    // get current state (read only)
    const CState* getCurrentState() const;

    // can we undo?
    bool canUndo();

    // can we redo?
    bool canRedo();

    // undo state
    void undo();

    // redo state
    void redo();


    // load
    void load(FILE* pFile);

    // save
    void save(FILE* pFile);

  protected:
    // name and state data
    CName *mName;
    CState* mState;

    // current state we're at
    CState* mCurrentState;

    // temporary string buffer
    char mStrBuffer[1024];

    // check track number
    bool checkTrack(int pTrack);

    // check lane number
    bool checkLane(int pLane);

    // convert pInt > -1 to string representation, <= -1 to blank
    char* toString(int pInt);
    char* toString2(int pInt);

    // logical clock for internal counting
    int getLogicalClock();

  public:
    // init model with test data
    void initWithTestData();

  private:
    // logical clock
    int mLogicalClock;

    // file for autosave
    char* mAutosaveFilename;
};

#endif // __model_h__
