// State.cpp: implementation of the CState class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "awana_grandprix_scoring.h"
#include "State.h"

#include <assert.h>

#include "Name.h"

#ifdef _DEBUG
#undef THIS_FILE
 static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define kMaxTracks 2
#define kMaxLanes 4

//////////////////////////////////////////////////////////////////////
// symbols
//////////////////////////////////////////////////////////////////////

#define kSymbolDivider      "-----"
#define kSymbolPool         "pool"
#define kSymbolScore        "score"
#define kSymbolSel1         "sel1"
#define kSymbolSel2         "sel2"
#define kSymbolTrack1Score  "track1score"
#define kSymbolTrack2Score  "track2score"

//////////////////////////////////////////////////////////////////////
// modes
//////////////////////////////////////////////////////////////////////

#define kModeDivider        0
#define kModePool           1
#define kModeScore          2
#define kModeSel1           3
#define kModeSel2           4
#define kModeTrack1Score    5
#define kModeTrack2Score    6

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CState::CState() {
  mLast = 0;
  mNext = 0;

  clearState();
}

CState::~CState() {
  if (mNext) {
    delete mNext;
    mNext = 0;
  }
}

//////////////////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////////////////

CState* CState::undo() {
  return mLast;
}

bool CState::canUndo() {
  return (mLast != 0);
}

CState* CState::redo() {
  return mNext;
}

bool CState::canRedo() {
  return (mNext != 0);
}

CState* CState::firstState() {
  while (mLast) {
    return mLast->firstState();
  }

  return this;
}

CState* CState::lastState() {
  while (mNext) {
    return mNext->lastState();
  }

  return this;
}

CState* CState::clone() {
  // create new state
  newState();

  // clone our own contents
  *mNext = *this;

  // fix up pointers
  mNext->mLast = this;
  mNext->mNext = 0;

  // return 'next' pointer
  return mNext;
}

void CState::initPoolWithCName(const CName& pNames) {
  clearState();

  for (CName::tIntStringMAP::const_iterator theIter = pNames.mNames.begin(); theIter != pNames.mNames.end(); theIter++) {
    mPool[theIter->first] = 0;
  }
}

bool CState::isClear() {
  bool theSelection0Empty = true, theSelection1Empty = true;
  bool theTrack0ScoreEmpty = true, theTrack1ScoreEmpty = true;

  {
    for (tIntIntMAP::const_iterator theIter = mSelection[0].begin(); theIter != mSelection[0].end(); theIter++) {
      if (theIter->second > -1) {
        theSelection0Empty = false;
        break;
      }
    }
  }

  {
    for (tIntIntMAP::const_iterator theIter = mSelection[1].begin(); theIter != mSelection[1].end(); theIter++) {
      if (theIter->second > -1) {
        theSelection1Empty = false;
        break;
      }
    }
  }

  {
    for (tIDScoreMAP::const_iterator theIter = mTrackScore[0].begin(); theIter != mTrackScore[0].end(); theIter++) {
      if (theIter->second.mID > -1) {
        theTrack0ScoreEmpty = false;
        break;
      }
    }
  }

  {
    for (tIDScoreMAP::const_iterator theIter = mTrackScore[1].begin(); theIter != mTrackScore[1].end(); theIter++) {
      if (theIter->second.mID > -1) {
        theTrack1ScoreEmpty = false;
        break;
      }
    }
  }

  return ((mNext == 0) &&
          (mPool.empty()) && (mScore.empty()) &&
          theSelection0Empty && theSelection1Empty &&
          theTrack0ScoreEmpty && theTrack1ScoreEmpty);
}

void CState::load(FILE* pFile) {
  // clear current state first..
  clearState();

  int theSel1Index = 0;
  int theSel2Index = 0;
  int theTrack1Index = 0;
  int theTrack2Index = 0;

  // now recursively go through the load process
  int theMode = -1;
  while (!feof(pFile)) {
    char theBuffer[1024];
    fgets(theBuffer, 1024, pFile);

    // identify
    if        (strncmp(theBuffer, kSymbolDivider, strlen(kSymbolDivider)) == 0) {
      theMode = kModeDivider;

      // recursive step
      newState();
      mNext->load(pFile);

      // check to see if anything has been loaded
      if (mNext->isClear()) {
        cleanupNext();
      }

    } else if (strncmp(theBuffer, kSymbolPool, strlen(kSymbolPool)) == 0) {
      theMode = kModePool;
    } else if (strncmp(theBuffer, kSymbolScore, strlen(kSymbolScore)) == 0) {
      theMode = kModeScore;
    } else if (strncmp(theBuffer, kSymbolSel1, strlen(kSymbolSel1)) == 0) {
      theMode = kModeSel1;
      theSel1Index = 0;
    } else if (strncmp(theBuffer, kSymbolSel2, strlen(kSymbolSel2)) == 0) {
      theMode = kModeSel2;
      theSel2Index = 0;
    } else if (strncmp(theBuffer, kSymbolTrack1Score, strlen(kSymbolTrack1Score)) == 0) {
      theMode = kModeTrack1Score;
      theTrack1Index = 0;
    } else if (strncmp(theBuffer, kSymbolTrack2Score, strlen(kSymbolTrack2Score)) == 0) {
      theMode = kModeTrack2Score;
      theTrack2Index = 0;
    } else {
      // process
      switch (theMode) {
        case kModeDivider :
          {
            // already processed...
          }; break;
        case kModePool :
          {
            int theID = -1, theTimesUsed = -1;
            sscanf(theBuffer, "%d %d\n", &theID, &theTimesUsed);
            mPool[theID] = theTimesUsed;
          }; break;
        case kModeScore :
          {
            char theSeps[] = ", ";
            char* theStr;
            int theID;

            theStr = strtok(theBuffer, theSeps);
            theID = atoi(theStr);

            // while there are tokens left, store them
            tIntLIST* theList = &(mScore[theID].mList);
            theStr = strtok(0, theSeps);
            while (theStr) {
              int theScore = atoi(theStr);
              theList->push_back(theScore);
              
              // iterative
              theStr = strtok(0, theSeps);
            } // while (theStr)
          }; break;
        case kModeSel1 :
          {
            mSelection[0][theSel1Index] = atoi(theBuffer);
            theSel1Index++;
          }; break;
        case kModeSel2 :
          {
            int theIndex = mSelection[1].size();
            mSelection[1][theSel2Index] = atoi(theBuffer);
            theSel2Index++;
          }; break;
        case kModeTrack1Score :
          {
            int theID = -1, theScore = -1;
            sscanf(theBuffer, "%d %d\n", &theID, &theScore);
            mTrackScore[0][theTrack1Index] = sIDScore(theID, theScore);
            theTrack1Index++;
          }; break;
        case kModeTrack2Score :
          {
            int theID = -1, theScore = -1;
            sscanf(theBuffer, "%d %d\n", &theID, &theScore);
            mTrackScore[1][theTrack2Index] = sIDScore(theID, theScore);
            theTrack2Index++;
          }; break;
      } // switch (theMode)
    } // if (strncmp(theBuffer, k..., strlen(k...)) == 0)
  } // while (!feof(pFile))
}

void CState::save(FILE* pFile) {
  // save all our previous states first
  if (mLast) {
    mLast->save(pFile);
  }

  // save my own state

  { // pool
    fprintf(pFile, "%s\n", kSymbolPool);
    for (tIntIntMAP::const_iterator theIter = mPool.begin(); theIter != mPool.end(); theIter++) {
      fprintf(pFile, "%d %d\n", theIter->first, theIter->second);
    }
  }

  { // score
    fprintf(pFile, "%s\n", kSymbolScore);
    for (tIntIntScoreListMAP::const_iterator theIter = mScore.begin(); theIter != mScore.end(); theIter++) {
      const tIntLIST* theList = &(theIter->second.mList);

      fprintf(pFile, "%d", theIter->first);

      for (tIntLIST::const_iterator theIter2 = theList->begin(); theIter2 != theList->end(); theIter2++) {
        fprintf(pFile, " %d", *theIter2);
      }

      fprintf(pFile, "\n");
    }
  }

  { // sel1
    //assert(mSelection[0].size() == 4);

    fprintf(pFile, "%s\n", kSymbolSel1);
    for (tIntIntMAP::const_iterator theIter = mSelection[0].begin(); theIter != mSelection[0].end(); theIter++) {
      fprintf(pFile, "%d\n", theIter->second);
    }
  }

  { // sel2
    //assert(mSelection[1].size() == 4);

    fprintf(pFile, "%s\n", kSymbolSel2);
    for (tIntIntMAP::const_iterator theIter = mSelection[1].begin(); theIter != mSelection[1].end(); theIter++) {
      fprintf(pFile, "%d\n", theIter->second);
    }
  }

  { // track1score
    //assert(mTrackScore[0].size() == 4);

    fprintf(pFile, "%s\n", kSymbolTrack1Score);
    for (tIDScoreMAP::const_iterator theIter = mTrackScore[0].begin(); theIter != mTrackScore[0].end(); theIter++) {
      fprintf(pFile, "%d %d\n", theIter->second.mID, theIter->second.mScore);
    }
  }
  
  { // track2score
    //assert(mTrackScore[1].size() == 4);

    fprintf(pFile, "%s\n", kSymbolTrack2Score);
    for (tIDScoreMAP::const_iterator theIter = mTrackScore[1].begin(); theIter != mTrackScore[1].end(); theIter++) {
      fprintf(pFile, "%d %d\n", theIter->second.mID, theIter->second.mScore);
    }
  }
  
  
  // output divider if we have a next state
  if (mNext) {
    fprintf(pFile, "%s\n", kSymbolDivider);
  }
}

//////////////////////////////////////////////////////////////////////
// protected
//////////////////////////////////////////////////////////////////////

void CState::cleanupNext() {
  if (mNext) {
    delete mNext;
    mNext = 0;
  }
}

void CState::clearState() {
  mPool.clear();
  mScore.clear();

  mSelection[0].clear();
  mSelection[1].clear();

  mTrackScore[0].clear();
  mTrackScore[1].clear();

  for (int i = 0; i < kMaxTracks; i++) {
    for (int j = 0; j < kMaxLanes; j++) {
      mSelection[i][j] = -1;
      mTrackScore[i][j] = CState::sIDScore(-1, 0);
    }
  }

  cleanupNext();
}

void CState::newState() {
  // cleanup 'next' pointer
  cleanupNext();

  // create new empty state
  CState* theNewState = new CState();

  // fix up pointers
  theNewState->mLast = this;
  mNext = theNewState;
}
